@@ -15,6 +15,7 @@ use crate::common::{
     AnimationControl, BitDepth, BlendOp, ColorType, DisposeOp, FrameControl, Info, ParameterError,
     PixelDimensions, ScaledFloat, SourceChromaticities, Unit,
 };
+use crate::text_metadata::{ITXtChunk, TEXtChunk, TextDecodingError, ZTXtChunk};
 use crate::traits::ReadBytesExt;
 
 /// TODO check if these size are reasonable
@@ -41,7 +42,7 @@ enum State {
     U32Byte2(U32Value, u32),
     U32Byte1(U32Value, u32),
     U32(U32Value),
-    ReadChunk(ChunkType, bool),
+    ReadChunk(ChunkType),
     PartialChunk(ChunkType),
     DecodeData(ChunkType, usize),
 }
@@ -124,6 +125,7 @@ pub(crate) enum FormatErrorInner {
         crc_val: u32,
         /// Calculated CRC32 sum
         crc_sum: u32,
+        /// The chunk type that has the CRC mismatch.
         chunk: ChunkType,
     },
     /// Not a PNG, the magic signature is missing.
@@ -152,12 +154,10 @@ pub(crate) enum FormatErrorInner {
         kind: ChunkType,
     },
     /// 4.3., some chunks must be between PLTE and IDAT.
-    // FIXME: why are we not using this?
     OutsidePlteIdat {
         kind: ChunkType,
     },
     /// 4.3., some chunks must be unique.
-    // FIXME: why are we not using this?
     DuplicateChunk {
         kind: ChunkType,
     },
@@ -178,8 +178,8 @@ pub(crate) enum FormatErrorInner {
     PaletteRequired,
     /// The color-depth combination is not valid according to Table 11.1.
     InvalidColorBitDepth {
-        color: ColorType,
-        depth: BitDepth,
+        color_type: ColorType,
+        bit_depth: BitDepth,
     },
     ColorWithBadTrns(ColorType),
     InvalidBitDepth(u8),
@@ -205,6 +205,8 @@ pub(crate) enum FormatErrorInner {
     // TODO: strictly type this.
     /// Filtering of a row has failed.
     BadFilter(&'static str),
+    /// Bad text encoding
+    BadTextEncoding(TextDecodingError),
 }
 
 impl error::Error for DecodingError {
@@ -233,11 +235,14 @@ impl fmt::Display for FormatError {
         use FormatErrorInner::*;
         match &self.inner {
             CrcMismatch {
-                crc_val, crc_sum, ..
+                crc_val,
+                crc_sum,
+                chunk,
+                ..
             } => write!(
                 fmt,
-                "CRC error: expected 0x{:x} have 0x{:x}",
-                crc_val, crc_sum
+                "CRC error: expected 0x{:x} have 0x{:x} while decoding {:?} chunk.",
+                crc_val, crc_sum, chunk
             ),
             MissingIhdr => write!(fmt, "IHDR chunk missing"),
             MissingFctl => write!(fmt, "fcTL chunk missing before fdAT chunk."),
@@ -262,10 +267,13 @@ impl fmt::Display for FormatError {
                 expected, len
             ),
             PaletteRequired => write!(fmt, "Missing palette of indexed image."),
-            InvalidColorBitDepth { color, depth } => write!(
+            InvalidColorBitDepth {
+                color_type,
+                bit_depth,
+            } => write!(
                 fmt,
                 "Invalid color/depth combination in header: {:?}/{:?}",
-                color, depth,
+                color_type, bit_depth,
             ),
             ColorWithBadTrns(color_type) => write!(
                 fmt,
@@ -286,9 +294,55 @@ impl fmt::Display for FormatError {
             UnexpectedEof => write!(fmt, "Unexpected end of data before image end."),
             UnexpectedEndOfChunk => write!(fmt, "Unexpected end of data within a chunk."),
             NoMoreImageData => write!(fmt, "IDAT or fDAT chunk is has not enough data for image."),
-            // TODO: figure out a good way to print the error.
-            CorruptFlateStream { err: _ } => write!(fmt, "Corrupt deflate stream."),
+            CorruptFlateStream { err } => {
+                write!(fmt, "Corrupt deflate stream. ")?;
+                use miniz_oxide::inflate::TINFLStatus;
+                match err {
+                    TINFLStatus::Adler32Mismatch => write!(fmt, "Adler32 checksum failed."),
+                    TINFLStatus::BadParam => write!(fmt, "Invalid input parameters."),
+                    // The Done status should already have been handled.
+                    TINFLStatus::Done => write!(fmt, "Unexpected done status."),
+                    TINFLStatus::FailedCannotMakeProgress => {
+                        write!(fmt, "Unexpected end of data.")
+                    }
+                    // The HasMoreOutput status should already have been handled.
+                    TINFLStatus::HasMoreOutput => write!(fmt, "Has more output."),
+                    // The HasMoreInput status should already have been handled.
+                    TINFLStatus::NeedsMoreInput => write!(fmt, "Needs more input."),
+                    // Write out the error number in case a new has been added.
+                    _ => write!(fmt, "Error number {:?}.", err),
+                }
+            }
             BadFilter(message) => write!(fmt, "{}.", message),
+            // TODO: Wrap more info in the enum variant
+            BadTextEncoding(tde) => {
+                match tde {
+                    TextDecodingError::Unrepresentable => {
+                        write!(fmt, "Unrepresentable data in tEXt chunk.")
+                    }
+                    TextDecodingError::InvalidKeywordSize => {
+                        write!(fmt, "Keyword empty or longer than 79 bytes.")
+                    }
+                    TextDecodingError::MissingNullSeparator => {
+                        write!(fmt, "No null separator in tEXt chunk.")
+                    }
+                    TextDecodingError::InflationError => {
+                        write!(fmt, "Invalid compressed text data.")
+                    }
+                    TextDecodingError::OutOfDecompressionSpace => {
+                        write!(fmt, "Out of decompression space. Try with a larger limit.")
+                    }
+                    TextDecodingError::InvalidCompressionMethod => {
+                        write!(fmt, "Using an unrecognized byte as compression method.")
+                    }
+                    TextDecodingError::InvalidCompressionFlag => {
+                        write!(fmt, "Using a flag that is not 0 or 255 as a compression flag for iTXt chunk.")
+                    }
+                    TextDecodingError::MissingCompressionFlag => {
+                        write!(fmt, "No compression flag in the iTXt chunk.")
+                    }
+                }
+            }
         }
     }
 }
@@ -320,6 +374,26 @@ impl From<DecodingError> for io::Error {
     }
 }
 
+impl From<TextDecodingError> for DecodingError {
+    fn from(tbe: TextDecodingError) -> Self {
+        DecodingError::Format(FormatError {
+            inner: FormatErrorInner::BadTextEncoding(tbe),
+        })
+    }
+}
+
+/// Decode config
+#[derive(Default)]
+pub(crate) struct DecodeConfig {
+    ignore_text_chunk: bool,
+}
+
+impl DecodeConfig {
+    pub fn set_ignore_text_chunk(&mut self, ignore_text_chunk: bool) {
+        self.ignore_text_chunk = ignore_text_chunk;
+    }
+}
+
 /// PNG StreamingDecoder (low-level interface)
 pub struct StreamingDecoder {
     state: Option<State>,
@@ -333,6 +407,7 @@ pub struct StreamingDecoder {
     /// Stores where in decoding an `fdAT` chunk we are.
     apng_seq_handled: bool,
     have_idat: bool,
+    decode_config: DecodeConfig,
 }
 
 struct ChunkState {
@@ -363,6 +438,7 @@ impl StreamingDecoder {
             current_seq_no: None,
             apng_seq_handled: false,
             have_idat: false,
+            decode_config: DecodeConfig::default(),
         }
     }
 
@@ -379,6 +455,20 @@ impl StreamingDecoder {
         self.have_idat = false;
     }
 
+    /// Provides access to the inner `info` field
+    pub fn info(&self) -> Option<&Info<'static>> {
+        self.info.as_ref()
+    }
+
+    /// Set decode config
+    pub(crate) fn set_decode_config(&mut self, decode_config: DecodeConfig) {
+        self.decode_config = decode_config;
+    }
+
+    pub fn set_ignore_text_chunk(&mut self, ignore_text_chunk: bool) {
+        self.decode_config.set_ignore_text_chunk(ignore_text_chunk);
+    }
+
     /// Low level StreamingDecoder interface.
     ///
     /// Allows to stream partial data to the encoder. Returns a tuple containing the bytes that have
@@ -477,8 +567,9 @@ impl StreamingDecoder {
                         self.current_chunk.crc.update(&type_str.0);
                         self.current_chunk.remaining = length;
                         self.apng_seq_handled = false;
+                        self.current_chunk.raw_bytes.clear();
                         goto!(
-                            ReadChunk(type_str, true),
+                            ReadChunk(type_str),
                             emit Decoded::ChunkBegin(length, type_str)
                         )
                     }
@@ -558,25 +649,31 @@ impl StreamingDecoder {
                             // complete chunk
                             Ok((0, self.parse_chunk(type_str)?))
                         } else {
+                            // Make sure we have room to read more of the chunk.
+                            // We need it fully before parsing.
+                            self.reserve_current_chunk()?;
+
                             goto!(
-                                0, ReadChunk(type_str, true),
+                                0, ReadChunk(type_str),
                                 emit Decoded::PartialChunk(type_str)
                             )
                         }
                     }
                 }
             }
-            ReadChunk(type_str, clear) => {
-                if clear {
-                    self.current_chunk.raw_bytes.clear();
-                }
-                if self.current_chunk.remaining > 0 {
+            ReadChunk(type_str) => {
+                // The _previous_ event wanted to return the contents of raw_bytes, and let the
+                // caller consume it,
+                if self.current_chunk.remaining == 0 {
+                    goto!(0, U32(U32Value::Crc(type_str)))
+                } else {
                     let ChunkState {
                         crc,
                         remaining,
                         raw_bytes,
                         type_: _,
                     } = &mut self.current_chunk;
+
                     let buf_avail = raw_bytes.capacity() - raw_bytes.len();
                     let bytes_avail = min(buf.len(), buf_avail);
                     let n = min(*remaining, bytes_avail as u32);
@@ -586,15 +683,14 @@ impl StreamingDecoder {
                         let buf = &buf[..n as usize];
                         crc.update(buf);
                         raw_bytes.extend_from_slice(buf);
+
                         *remaining -= n;
                         if *remaining == 0 {
                             goto!(n as usize, PartialChunk(type_str))
                         } else {
-                            goto!(n as usize, ReadChunk(type_str, false))
+                            goto!(n as usize, ReadChunk(type_str))
                         }
                     }
-                } else {
-                    goto!(0, U32(U32Value::Crc(type_str)))
                 }
             }
             DecodeData(type_str, mut n) => {
@@ -603,9 +699,10 @@ impl StreamingDecoder {
                 let c = self.inflater.decompress(chunk_data, image_data)?;
                 n += c;
                 if n == chunk_len && c == 0 {
+                    self.current_chunk.raw_bytes.clear();
                     goto!(
                         0,
-                        ReadChunk(type_str, true),
+                        ReadChunk(type_str),
                         emit Decoded::ImageData
                     )
                 } else {
@@ -619,6 +716,22 @@ impl StreamingDecoder {
         }
     }
 
+    fn reserve_current_chunk(&mut self) -> Result<(), DecodingError> {
+        // FIXME: use limits, also do so in iccp/zlib decompression.
+        const MAX: usize = 0x10_0000;
+        let ref mut buffer = self.current_chunk.raw_bytes;
+
+        // Double if necessary, but no more than until the limit is reached.
+        let reserve_size = MAX.saturating_sub(buffer.capacity()).min(buffer.len());
+        buffer.reserve_exact(reserve_size);
+
+        if buffer.capacity() == buffer.len() {
+            Err(DecodingError::LimitsExceeded)
+        } else {
+            Ok(())
+        }
+    }
+
     fn parse_chunk(&mut self, type_str: ChunkType) -> Result<Decoded, DecodingError> {
         self.state = Some(State::U32(U32Value::Crc(type_str)));
         if self.info.is_none() && type_str != IHDR {
@@ -637,6 +750,9 @@ impl StreamingDecoder {
             chunk::cHRM => self.parse_chrm(),
             chunk::sRGB => self.parse_srgb(),
             chunk::iCCP => self.parse_iccp(),
+            chunk::tEXt if !self.decode_config.ignore_text_chunk => self.parse_text(),
+            chunk::zTXt if !self.decode_config.ignore_text_chunk => self.parse_ztxt(),
+            chunk::iTXt if !self.decode_config.ignore_text_chunk => self.parse_itxt(),
             _ => Ok(Decoded::PartialChunk(type_str)),
         } {
             Err(err) => {
@@ -648,12 +764,6 @@ impl StreamingDecoder {
         }
     }
 
-    fn get_info_or_err(&self) -> Result<&Info, DecodingError> {
-        self.info
-            .as_ref()
-            .ok_or_else(|| DecodingError::Format(FormatErrorInner::MissingIhdr.into()))
-    }
-
     fn parse_fctl(&mut self) -> Result<Decoded, DecodingError> {
         let mut buf = &self.current_chunk.raw_bytes[..];
         let next_seq_no = buf.read_be()?;
@@ -736,27 +846,28 @@ impl StreamingDecoder {
     }
 
     fn parse_plte(&mut self) -> Result<Decoded, DecodingError> {
-        if let Some(info) = self.info.as_mut() {
-            info.palette = Some(Cow::Owned(self.current_chunk.raw_bytes.clone()))
+        let info = self.info.as_mut().unwrap();
+        if info.palette.is_some() {
+            // Only one palette is allowed
+            return Err(DecodingError::Format(
+                FormatErrorInner::DuplicateChunk { kind: chunk::PLTE }.into(),
+            ));
+        } else {
+            info.palette = Some(Cow::Owned(self.current_chunk.raw_bytes.clone()));
+            Ok(Decoded::Nothing)
         }
-        Ok(Decoded::Nothing)
     }
 
     fn parse_trns(&mut self) -> Result<Decoded, DecodingError> {
-        let (color_type, bit_depth) = {
-            let info = self.get_info_or_err()?;
-            (info.color_type, info.bit_depth as u8)
-        };
+        let info = self.info.as_mut().unwrap();
+        if info.trns.is_some() {
+            return Err(DecodingError::Format(
+                FormatErrorInner::DuplicateChunk { kind: chunk::PLTE }.into(),
+            ));
+        }
+        let (color_type, bit_depth) = { (info.color_type, info.bit_depth as u8) };
         let mut vec = self.current_chunk.raw_bytes.clone();
         let len = vec.len();
-        let info = match self.info {
-            Some(ref mut info) => info,
-            None => {
-                return Err(DecodingError::Format(
-                    FormatErrorInner::ChunkBeforeIhdr { kind: chunk::tRNS }.into(),
-                ))
-            }
-        };
         match color_type {
             ColorType::Grayscale => {
                 if len < 2 {
@@ -787,10 +898,18 @@ impl StreamingDecoder {
                 Ok(Decoded::Nothing)
             }
             ColorType::Indexed => {
-                // FIXME: what's going on here??
-                let _ = info.palette.as_ref().ok_or_else(|| {
-                    DecodingError::Format(FormatErrorInner::AfterPlte { kind: chunk::tRNS }.into())
-                });
+                // The transparency chunk must be after the palette chunk and
+                // before the data chunk.
+                if info.palette.is_none() {
+                    return Err(DecodingError::Format(
+                        FormatErrorInner::AfterPlte { kind: chunk::tRNS }.into(),
+                    ));
+                } else if self.have_idat {
+                    return Err(DecodingError::Format(
+                        FormatErrorInner::OutsidePlteIdat { kind: chunk::tRNS }.into(),
+                    ));
+                }
+
                 info.trns = Some(Cow::Owned(vec));
                 Ok(Decoded::Nothing)
             }
@@ -801,10 +920,15 @@ impl StreamingDecoder {
     }
 
     fn parse_phys(&mut self) -> Result<Decoded, DecodingError> {
+        let info = self.info.as_mut().unwrap();
         if self.have_idat {
             Err(DecodingError::Format(
                 FormatErrorInner::AfterIdat { kind: chunk::pHYs }.into(),
             ))
+        } else if info.pixel_dims.is_some() {
+            return Err(DecodingError::Format(
+                FormatErrorInner::DuplicateChunk { kind: chunk::pHYs }.into(),
+            ));
         } else {
             let mut buf = &self.current_chunk.raw_bytes[..];
             let xppu = buf.read_be()?;
@@ -819,17 +943,22 @@ impl StreamingDecoder {
                 }
             };
             let pixel_dims = PixelDimensions { xppu, yppu, unit };
-            self.info.as_mut().unwrap().pixel_dims = Some(pixel_dims);
+            info.pixel_dims = Some(pixel_dims);
             Ok(Decoded::PixelDimensions(pixel_dims))
         }
     }
 
     fn parse_chrm(&mut self) -> Result<Decoded, DecodingError> {
+        let info = self.info.as_mut().unwrap();
         if self.have_idat {
             Err(DecodingError::Format(
-                FormatErrorInner::AfterIdat { kind: chunk::gAMA }.into(),
+                FormatErrorInner::AfterIdat { kind: chunk::cHRM }.into(),
             ))
-        } else if self.info.as_ref().unwrap().srgb.is_some() {
+        } else if info.source_chromaticities.is_some() {
+            return Err(DecodingError::Format(
+                FormatErrorInner::DuplicateChunk { kind: chunk::cHRM }.into(),
+            ));
+        } else if info.srgb.is_some() {
             // Ignore chromaticities if sRGB profile is used.
             Ok(Decoded::Nothing)
         } else {
@@ -862,27 +991,23 @@ impl StreamingDecoder {
                 ),
             };
 
-            let info = match self.info {
-                Some(ref mut info) => info,
-                None => {
-                    return Err(DecodingError::Format(
-                        FormatErrorInner::ChunkBeforeIhdr { kind: chunk::tRNS }.into(),
-                    ))
-                }
-            };
-
             info.source_chromaticities = Some(source_chromaticities);
             Ok(Decoded::Nothing)
         }
     }
 
     fn parse_gama(&mut self) -> Result<Decoded, DecodingError> {
+        let info = self.info.as_mut().unwrap();
         if self.have_idat {
             Err(DecodingError::Format(
                 FormatErrorInner::AfterIdat { kind: chunk::gAMA }.into(),
             ))
-        } else if self.info.as_ref().unwrap().srgb.is_some() {
-            // Ignore gamma data if sRGB profile is used.
+        } else if info.source_gamma.is_some() {
+            return Err(DecodingError::Format(
+                FormatErrorInner::DuplicateChunk { kind: chunk::gAMA }.into(),
+            ));
+        } else if info.srgb.is_some() {
+            // Ignore chromaticities if sRGB profile is used.
             Ok(Decoded::Nothing)
         } else {
             let mut buf = &self.current_chunk.raw_bytes[..];
@@ -893,10 +1018,15 @@ impl StreamingDecoder {
     }
 
     fn parse_srgb(&mut self) -> Result<Decoded, DecodingError> {
+        let info = self.info.as_mut().unwrap();
         if self.have_idat {
             Err(DecodingError::Format(
                 FormatErrorInner::AfterIdat { kind: chunk::acTL }.into(),
             ))
+        } else if info.srgb.is_some() {
+            return Err(DecodingError::Format(
+                FormatErrorInner::DuplicateChunk { kind: chunk::sRGB }.into(),
+            ));
         } else {
             let mut buf = &self.current_chunk.raw_bytes[..];
             let raw: u8 = buf.read_be()?; // BE is is nonsense for single bytes, but this way the size is checked.
@@ -905,7 +1035,6 @@ impl StreamingDecoder {
             })?;
 
             // Set srgb and override source gamma and chromaticities.
-            let info = self.info.as_mut().unwrap();
             info.srgb = Some(rendering_intent);
             info.source_gamma = Some(crate::srgb::substitute_gamma());
             info.source_chromaticities = Some(crate::srgb::substitute_chromaticities());
@@ -914,10 +1043,15 @@ impl StreamingDecoder {
     }
 
     fn parse_iccp(&mut self) -> Result<Decoded, DecodingError> {
+        let info = self.info.as_mut().unwrap();
         if self.have_idat {
             Err(DecodingError::Format(
                 FormatErrorInner::AfterIdat { kind: chunk::iCCP }.into(),
             ))
+        } else if info.icc_profile.is_some() {
+            return Err(DecodingError::Format(
+                FormatErrorInner::DuplicateChunk { kind: chunk::iCCP }.into(),
+            ));
         } else {
             let mut buf = &self.current_chunk.raw_bytes[..];
 
@@ -951,13 +1085,17 @@ impl StreamingDecoder {
                 buf = &buf[consumed_bytes..];
             }
 
-            self.info.as_mut().unwrap().icc_profile = Some(Cow::Owned(profile));
+            info.icc_profile = Some(Cow::Owned(profile));
             Ok(Decoded::Nothing)
         }
     }
 
     fn parse_ihdr(&mut self) -> Result<Decoded, DecodingError> {
-        // TODO: check if color/bit depths combination is valid
+        if self.info.is_some() {
+            return Err(DecodingError::Format(
+                FormatErrorInner::DuplicateChunk { kind: IHDR }.into(),
+            ));
+        }
         let mut buf = &self.current_chunk.raw_bytes[..];
         let width = buf.read_be()?;
         let height = buf.read_be()?;
@@ -972,7 +1110,19 @@ impl StreamingDecoder {
         };
         let color_type = buf.read_be()?;
         let color_type = match ColorType::from_u8(color_type) {
-            Some(color_type) => color_type,
+            Some(color_type) => {
+                if color_type.is_combination_invalid(bit_depth) {
+                    return Err(DecodingError::Format(
+                        FormatErrorInner::InvalidColorBitDepth {
+                            color_type,
+                            bit_depth,
+                        }
+                        .into(),
+                    ));
+                } else {
+                    color_type
+                }
+            }
             None => {
                 return Err(DecodingError::Format(
                     FormatErrorInner::InvalidColorType(color_type).into(),
@@ -1006,18 +1156,113 @@ impl StreamingDecoder {
                 ))
             }
         };
-        let mut info = Info::default();
 
-        info.width = width;
-        info.height = height;
-        info.bit_depth = bit_depth;
-        info.color_type = color_type;
-        info.interlaced = interlaced;
-        self.info = Some(info);
+        self.info = Some(Info {
+            width,
+            height,
+            bit_depth,
+            color_type,
+            interlaced,
+            ..Default::default()
+        });
+
         Ok(Decoded::Header(
             width, height, bit_depth, color_type, interlaced,
         ))
     }
+
+    fn split_keyword(buf: &[u8]) -> Result<(&[u8], &[u8]), DecodingError> {
+        let null_byte_index = buf
+            .iter()
+            .position(|&b| b == 0)
+            .ok_or_else(|| DecodingError::from(TextDecodingError::MissingNullSeparator))?;
+
+        if null_byte_index == 0 || null_byte_index > 79 {
+            return Err(DecodingError::from(TextDecodingError::InvalidKeywordSize));
+        }
+
+        Ok((&buf[..null_byte_index], &buf[null_byte_index + 1..]))
+    }
+
+    fn parse_text(&mut self) -> Result<Decoded, DecodingError> {
+        let buf = &self.current_chunk.raw_bytes[..];
+
+        let (keyword_slice, value_slice) = Self::split_keyword(buf)?;
+
+        self.info
+            .as_mut()
+            .unwrap()
+            .uncompressed_latin1_text
+            .push(TEXtChunk::decode(keyword_slice, value_slice).map_err(DecodingError::from)?);
+
+        Ok(Decoded::Nothing)
+    }
+
+    fn parse_ztxt(&mut self) -> Result<Decoded, DecodingError> {
+        let buf = &self.current_chunk.raw_bytes[..];
+
+        let (keyword_slice, value_slice) = Self::split_keyword(buf)?;
+
+        let compression_method = *value_slice
+            .get(0)
+            .ok_or_else(|| DecodingError::from(TextDecodingError::InvalidCompressionMethod))?;
+
+        let text_slice = &value_slice[1..];
+
+        self.info.as_mut().unwrap().compressed_latin1_text.push(
+            ZTXtChunk::decode(keyword_slice, compression_method, text_slice)
+                .map_err(DecodingError::from)?,
+        );
+
+        Ok(Decoded::Nothing)
+    }
+
+    fn parse_itxt(&mut self) -> Result<Decoded, DecodingError> {
+        let buf = &self.current_chunk.raw_bytes[..];
+
+        let (keyword_slice, value_slice) = Self::split_keyword(buf)?;
+
+        let compression_flag = *value_slice
+            .get(0)
+            .ok_or_else(|| DecodingError::from(TextDecodingError::MissingCompressionFlag))?;
+
+        let compression_method = *value_slice
+            .get(1)
+            .ok_or_else(|| DecodingError::from(TextDecodingError::InvalidCompressionMethod))?;
+
+        let second_null_byte_index = value_slice[2..]
+            .iter()
+            .position(|&b| b == 0)
+            .ok_or_else(|| DecodingError::from(TextDecodingError::MissingNullSeparator))?
+            + 2;
+
+        let language_tag_slice = &value_slice[2..second_null_byte_index];
+
+        let third_null_byte_index = value_slice[second_null_byte_index + 1..]
+            .iter()
+            .position(|&b| b == 0)
+            .ok_or_else(|| DecodingError::from(TextDecodingError::MissingNullSeparator))?
+            + (second_null_byte_index + 1);
+
+        let translated_keyword_slice =
+            &value_slice[second_null_byte_index + 1..third_null_byte_index];
+
+        let text_slice = &value_slice[third_null_byte_index + 1..];
+
+        self.info.as_mut().unwrap().utf8_text.push(
+            ITXtChunk::decode(
+                keyword_slice,
+                compression_flag,
+                compression_method,
+                language_tag_slice,
+                translated_keyword_slice,
+                text_slice,
+            )
+            .map_err(DecodingError::from)?,
+        );
+
+        Ok(Decoded::Nothing)
+    }
 }
 
 impl Info<'_> {
